/* indent:4  tabsize:8  font:fixed-width */

#import "CardPile.h"

@implementation CardPile


/*---------------------------------------------------------------------------
|
|    + initialize
|
|    returns: (id) [super initialize]
|
|----------------------------------------------------------------------------
|
|    Sets a random seed based on the time before initializing the class
|            
\----------------------------------------------------------------------------*/

+ initialize
{
    srandom(time(0));
    return [super initialize];
}


/*---------------------------------------------------------------------------
|
|    - init
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Creates an empty list to manage the cards in the pile after 
|    initializing the object
|            
\----------------------------------------------------------------------------*/

- init
{
    [self initForCardSize:CS_SMALL];
    return self;
}


/*---------------------------------------------------------------------------
|
|    - initForCardSize:(CardSize)aSize
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Designated initializer for CardPile class.
|
\----------------------------------------------------------------------------*/

- initForCardSize:(CardSize)aSize
{
    [super init];
    cardSize = aSize;
	numDecks = 0;
    cardList = [[List allocFromZone:[self zone]]init];
    return self;
}


/*---------------------------------------------------------------------------
|
|    - setCardSize:(CardSize)aSize
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Set the card size.
|
\----------------------------------------------------------------------------*/

- setCardSize:(CardSize)aSize
{
    if (aSize != cardSize)
    {
        // have to change all the cards to the right class
	
        id newCard;
	id oldCard;
	int i;
	int cardCount = [cardList count];
	
        for (i = 0; i < cardCount; i++)
	{
	    oldCard = [cardList objectAt:i];
	    if (aSize == CS_SMALL)
	    {
	        newCard = [[SmallCard allocFromZone:[self zone]]
		              initSuit:[oldCard suit]
		                 value:[oldCard value]
				 faceUp:[oldCard isFaceUp]];
            }
	    else
	    {
	        newCard = [[Card allocFromZone:[self zone]] 
			      initSuit:[oldCard suit]
		                 value:[oldCard value]
				 faceUp:[oldCard isFaceUp]];
	    }
	    [cardList removeObjectAt:i];
	    [oldCard free];
	    [cardList insertObject:newCard at:i];

	}
        cardSize = aSize;
    }
    return self;
}


/*---------------------------------------------------------------------------
|
|    - (CardSize)returnSize
|
|    returns: (CardSize) current card size
|
|----------------------------------------------------------------------------
|
|    Return the current card size.
|
\----------------------------------------------------------------------------*/

- (CardSize)returnSize
{
    return cardSize;
}


/*---------------------------------------------------------------------------
|
|    - free
|
|    returns: (id) [super free]
|
|----------------------------------------------------------------------------
|
|    Discards all of the card objects, the list they are contained in, 
|    and finally the CardPile itself.
|            
\----------------------------------------------------------------------------*/

- free
{
    [self freeCards];
    [cardList free];
    return [super free];
}


/*---------------------------------------------------------------------------
|
|    - addCard:aCard
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Acts as a cover for insertCard::, but always adds cards to the top of 
|    a pile.
|
\----------------------------------------------------------------------------*/

- addCard:aCard
{
    return [self insertCard:aCard at:CS_TOP];
}


/*---------------------------------------------------------------------------
|
|    - insertCard:aCard at:(int) aPosition
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Adds a card to the pile at a specific position.  The constants CS_TOP 
|    and CS_BOTTOM can be used to specify the top and bottom of the pile, 
|    respectively.
|
\----------------------------------------------------------------------------*/

- insertCard:aCard at:(int)aPosition
{
    if (aPosition == CS_TOP)
    {
        [cardList addObject:aCard];
    }
    else
    {
        [cardList insertObject:aCard at:aPosition];
    }
    return self;
}


/*---------------------------------------------------------------------------
|
|    - addPile:aCardPile
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Acts as a cover for insertCardPile::, but always adds cards to the top 
|    of a pile.
|
\----------------------------------------------------------------------------*/

- addPile:aCardPile
{
    return [self insertPile:aCardPile at:CS_TOP];
}


/*---------------------------------------------------------------------------
|
|    - addCopyOfPile:aCardPile
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Acts as a cover for insertCopyOfCardPile:at:, but always adds cards to 
|    the top of a pile.
|
\----------------------------------------------------------------------------*/

- addCopyOfPile:aCardPile
{
    return [self insertCopyOfPile:aCardPile at:CS_TOP];
}


/*---------------------------------------------------------------------------
|
|    - insertPile:aCardPile at:(int)aPosition
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Adds another pile of cards to this pile at a specific position.  The
|    constants CS_TOP and CS_BOTTOM can be used to specify the top and bottom
|    of the pile, respectively.
|
\----------------------------------------------------------------------------*/

- insertPile:aCardPile at:(int)aPosition
{
    int        cardIndex;

    if (aPosition == CS_TOP)
    {
        for (cardIndex = 0; cardIndex < [aCardPile cardCount]; cardIndex++)
        {
            [cardList addObject:[aCardPile cardAt:cardIndex]];
        }
    }
    else
    {
        for (cardIndex = 0; cardIndex < [aCardPile cardCount]; cardIndex++)
        {
            [cardList insertObject:[aCardPile cardAt:cardIndex]at:aPosition++];
        }
    }
    return self;
}

/*---------------------------------------------------------------------------
|
|    - insertCopyOfPile:aCardPile at:(int)aPosition
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Adds another pile of cards to this pile at a specific position.  The
|    constants CS_TOP and CS_BOTTOM can be used to specify the top and bottom
|    of the pile, respectively.  The cards added are copies of (and 
|    independent from) the source cards.
|
\----------------------------------------------------------------------------*/

- insertCopyOfPile:aCardPile at:(int)aPosition
{
    int        cardIndex;

    if (aPosition == CS_TOP)
    {
        for (cardIndex = 0; cardIndex < [aCardPile cardCount]; cardIndex++)
        {
            [cardList addObject:[[aCardPile cardAt:cardIndex]copy]];
        }
    }
    else
    {
        for (cardIndex = 0; cardIndex < [aCardPile cardCount]; cardIndex++)
        {
            [cardList insertObject:[[aCardPile cardAt:cardIndex]copy] 
	                        at:aPosition++];
        }
    }
    return self;
}

/*---------------------------------------------------------------------------
|
|    - removeCard:aCard
|
|    returns: (id) nil if the object wasn't in the CardPile
|
|             (id) aCard if the card was removed
|
|----------------------------------------------------------------------------
|
|    Remove a specific card from the pile.  The card is not deallocated, it 
|    is up to the client to reuse or free the object.  Note that this is the 
|    only way to remove a card from the pile.
|    
\----------------------------------------------------------------------------*/

- removeCard:aCard
{
    return [cardList removeObject:aCard];
}


/*---------------------------------------------------------------------------
|
|    - addDeck
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Add a full deck of cards to the top of the pile, neatly ordered.
|    
\----------------------------------------------------------------------------*/

- addDeck
{
    int suitCounter, valueCounter;

    for (valueCounter = CS_LOWVALUE; 
         valueCounter <= CS_HIGHVALUE;  
	 valueCounter++)
    {
        for (suitCounter = CS_LOWSUIT; 
	     suitCounter <= CS_HIGHSUIT;  
	     suitCounter++)
        {
            if (cardSize == CS_SMALL)
            {
                [self insertCard:[[SmallCard allocFromZone:[self zone]]
		                                  initSuit:suitCounter 
		                                     value:valueCounter]
                                                        at:CS_TOP];
            }
            else if (cardSize == CS_LARGE)
            {
                [self insertCard:[[Card allocFromZone:[self zone]]
		                             initSuit:suitCounter 
		                                value:valueCounter]
                                                   at:CS_TOP];
            }
        }
    }
	
	numDecks++;
	
    return self;
}

/*---------------------------------------------------------------------------
|
|    - empty
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Empties the card list describing the pile.
|    
\----------------------------------------------------------------------------*/

- empty
{
    [cardList empty];
	numDecks = 0;
	
    return self;
}


/*---------------------------------------------------------------------------
|
|    - freeCards
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Discards all of the card objects contained in the pile.
|    
\----------------------------------------------------------------------------*/

- freeCards
{
    [cardList freeObjects];
    return self;
}


/*---------------------------------------------------------------------------
|
|    - flip
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Turns the pile over, reversing the order of the cards and the 
|    orientation of each card.
|
\----------------------------------------------------------------------------*/

- flip
{
    int cardIndex;

    for (cardIndex = [cardList count]- 1; cardIndex >= 0; cardIndex--)
    {
        [cardList addObject:[[cardList removeObjectAt:cardIndex]flip]];
    }
    return self;
}



/*---------------------------------------------------------------------------
|
|    - shuffle
|
|    returns: (id) self
|
|----------------------------------------------------------------------------
|
|    Randomly re-order all card objects contained in the pile.
|    
\----------------------------------------------------------------------------*/

- shuffle
{
    id         temp;
    int        counter;
    int        swapIndex;
    int        cardCount = [cardList count];

    for (counter = 0; counter < cardCount - 1; counter++)
    {
        swapIndex = ( random () % ( cardCount - counter)) + counter;
        temp = [cardList objectAt:counter];
        [cardList replaceObjectAt:counter with:[cardList objectAt:swapIndex]];
        [cardList replaceObjectAt:swapIndex with:temp];
    }
    return self;
}


/*---------------------------------------------------------------------------
|
|    - cardAt:(int) aPosition
|
|    returns: (id)        A card object for the given index, if it exists
|
|             (id)        nil if there is no card at the specified index
|
|----------------------------------------------------------------------------
|
|    Allow a client to find out what card is at a given position in the pile.  
|    The constants CS_TOP and CS_BOTTOM can be used to specify the top and 
|    bottom of the pile, respectively.
|    
\----------------------------------------------------------------------------*/

- cardAt:(int)aPosition
{
    if (aPosition == CS_TOP)
    {
        aPosition = [cardList count]- 1;
        if (aPosition == -1)
        {
            return nil;
        }
    }
    return [cardList objectAt:aPosition];
}


/*---------------------------------------------------------------------------
|
|    - (int) cardIndex:aCard
|
|    returns: (int)        A index for the given card, if it exists in the pile
|
|             (int)        NX_NOT_IN_LIST if the card isn't in the pile
|
|----------------------------------------------------------------------------
|
|    Allow a client to find out what card is at a given position in the stack.
|    
\----------------------------------------------------------------------------*/

- (int)cardIndex:aCard
{
    return [cardList indexOf:aCard];
}


/*---------------------------------------------------------------------------
|
|    - (int) cardCount
|
|    returns: (int) The number of cards in the pile
|
|----------------------------------------------------------------------------
|
|    Allow a client to find out how many cards are in the pile.
|    
\----------------------------------------------------------------------------*/

- (int) cardCount
{
    return [cardList count];
}

/*---------------------------------------------------------------------------
|
|    - read:(NXTypedStream *) theStream
|
|    returns: (int) self
|
|----------------------------------------------------------------------------
|
|    Dearchive a CardPile.
|    
\----------------------------------------------------------------------------*/

- read:(NXTypedStream *) theStream
{
    [super read:theStream];
    cardList = NXReadObject(theStream);
    NXReadTypes(theStream, "i", &cardSize);
    return self;
}


/*---------------------------------------------------------------------------
|
|    - write:(NXTypedStream *) theStream
|
|    returns: (int) self
|
|----------------------------------------------------------------------------
|
|    Archive a CardPile.
|    
\----------------------------------------------------------------------------*/

- write:(NXTypedStream *)theStream
{
    [super write:theStream];
    NXWriteObject(theStream, cardList);
    NXWriteTypes(theStream, "i", &cardSize);
    return self;
}

/*******************************************************************************
 *                                                                             *
 *                      Stuff added by Mark Trombino                           *     
 *                                                                             *
 *******************************************************************************/
- addDecks:(int)numdecks
/*
	Adds numDecks to our pile
*/
{
	int i;
	
	for(i=0; i<numdecks; i++)
		[self addDeck];
		
	return self;
}

- (int) deckCount
{
	return numDecks;
}

- (int)total
/*
	returns the total of all the values of the cards in our pile
*/
{
	int i, total = 0;
	
	for(i=0; i<[cardList count]; i++)
		total += [[cardList objectAt:i] value];
	
	return total;
}

- (int)userTotal
/*
	Returns the total of all the userValues of the cards in our pile
*/
{
	int i, total = 0;
	
	for(i=0; i<[cardList count]; i++)
		total += [[cardList objectAt:i] userValue];
	
	return total;
}

- setAceHigh:(BOOL)flag
{
	aceHigh = flag;
	
	return self;
}

- (BOOL) isJacksOrBetter
/*
	Returns YES if there's at least one pair with jacks or better.
*/
{
	int i;
	
	for(i=CS_JACK; i<=CS_KING; i++)
	{
		if([self numCardsOfValue:i] == 2)
			return YES;
	}
	
	if(aceHigh && [self numCardsOfValue:CS_ACE] == 2)
		return YES;
		
	return NO;
}

- (BOOL) isTwoPair
/*
	Returns YES if there's at least two pairs in cardList.
	
	Since this message is called after we've checked for three of a kind, I cut
	some corners here...
*/
{
	int i, numPairs=0;
	
	for(i=CS_LOWVALUE; i<=CS_HIGHVALUE; i++)
	{
		if([self numCardsOfValue:i] == 2)
			numPairs++;
	}

	if(numPairs == 2)
		return YES;

	return NO;
}

- (BOOL) isThreeOfKind
/*
	Returns YES if cardList contains three of a kind.
*/
{
	int i;
	
	for(i=CS_LOWVALUE; i<=CS_HIGHVALUE; i++)
	{
		if([self numCardsOfValue:i] == 3)
			return YES;
	}
	
	return NO;
}
 
- (BOOL) isStraight
/*
	Returns YES if cardList is a straight hand.
*/
{
	int i, thisValue, nextValue;
	BOOL isStraight=YES;
	id sortedList = [cardList copyFromZone:[self zone]];
	
	sortedList = [self sortList:sortedList];
	
	for(i=0; i<[sortedList count]-1; i++)
	{
		thisValue = [[sortedList objectAt:i] value];
		if(thisValue == CS_ACE && aceHigh)
			thisValue = CS_KING+1;
		nextValue = [[sortedList objectAt:i+1] value];
		if(nextValue == CS_ACE && aceHigh)
			nextValue = CS_KING+1;
		
		if((nextValue - thisValue) != 1)
			return NO;
				
	}
	
	return isStraight;
}

- (BOOL) isFlush
/*
	Returns YES if cardList has all the same suit.
*/
{
	int i, suit;
	
	suit = [[cardList objectAt:0] suit];
	
	for(i=1; i<[cardList count]; i++)
	{
		if([[cardList objectAt:i] suit] != suit)
			return NO;
	}
	
	return YES;
}

- (BOOL) isFullHouse
/*
	Returns YES if cardList has one three of a kind and one pair.
*/
{
	int i, j;
	
	for(i=CS_LOWVALUE; i<=CS_HIGHVALUE; i++)
	{
		if([self numCardsOfValue:i] == 3)
		{
			for(j=CS_LOWVALUE; j<=CS_HIGHVALUE; j++)
			{
				if(j != i)
				{
					if([self numCardsOfValue:j] == 2)
						
						return YES;
				}
			}
		}
	}

	return NO;
}

- (BOOL) isFourOfKind
/*
	Returns YES if cardList contains four of a kind.
*/
{
	int i;
	
	for(i=CS_LOWVALUE; i<=CS_HIGHVALUE; i++)
	{
		if([self numCardsOfValue:i] == 4)
			return YES;
	}

	return NO;
}

- (BOOL)isStraightFlush
/*
	return YES if cardList is a straight flush.
*/
{
	
	if([self isStraight] && [self isFlush])
		return YES;
		
	return NO;
}

- (BOOL)isRoyalFlush
/*
	return YES if cardList is a royal flush.
*/
{
	
	if([self isStraight] && [self isFlush] && ([self highestValue] == CS_ACE) && aceHigh)
		return YES;
		
	return NO;
}

- (BOOL)isBlackjack
{
	if([self cardCount] == 2 && [self numCardsOfValue:CS_ACE] == 1 &&
		([self numCardsOfValue:CS_KING] == 1 || [self numCardsOfValue:CS_QUEEN] == 1 ||
		 [self numCardsOfValue:CS_JACK] == 1 || [self numCardsOfValue:CS_TEN] == 1))
		 	return YES;
	
	return NO;
}

- (int)numCardsOfValue:(int)aValue
/*
	Scans cardList to see how many cards have that userValue.  Returns that amount.
*/
{
	int i, count = 0;
	
	for(i=0; i<[cardList count]; i++)
	{
		if([[cardList objectAt:i] value] == aValue)
			count++;
	}
	
	return count;
}

- (int)lowestValue
{
	int i, value, thatValue;
	
	value = [[cardList objectAt:0] value];
	if(value == CS_ACE && aceHigh)
		value = CS_KING+1;
	
	for(i=1; i<[cardList count]; i++)
	{
		thatValue = [[cardList objectAt:i] value];
		if(thatValue == CS_ACE && aceHigh)
			thatValue = CS_KING+1;
		if(thatValue < value)
			value = thatValue;
	}
	
	if(value == CS_KING+1)
		value = CS_ACE;
	
	return value;
}

- lowestCard
{
	int i, value, thatValue;
	id card;
	
	value = [[cardList objectAt:0] value];
	if(value == CS_ACE && aceHigh)
		value = CS_KING+1;
	card = [cardList objectAt:0];
	
	for(i=1; i<[cardList count]; i++)
	{
		thatValue = [[cardList objectAt:i] value];
		if(thatValue == CS_ACE && aceHigh)
			thatValue = CS_KING+1;
		if(thatValue < value)
		{
			value = thatValue;
			card = [cardList objectAt:i];
		}
	}
	
	return card;
}

- (int)highestValue
{
	int i, value, thatValue;
	
	value = [[cardList objectAt:0] value];
	if(value == CS_ACE && aceHigh)
		value = CS_KING+1;
	
	for(i=1; i<[cardList count]; i++)
	{
		thatValue = [[cardList objectAt:i] value];
		if(thatValue == CS_ACE && aceHigh)
			thatValue = CS_KING+1;
		if(thatValue > value)
			value = thatValue;
	}
	
	if(value == CS_KING+1)
		value = CS_ACE;
	
	return value;
}

- highestCard
{
	int i, value, thatValue;
	id card;
	 
	value = [[cardList objectAt:0] value];
	if(value == CS_ACE && aceHigh)
		value = CS_KING+1;
	card = [cardList objectAt:0];
	
	for(i=1; i<[cardList count]; i++)
	{
		thatValue = [[cardList objectAt:i] value];
		if(thatValue == CS_ACE && aceHigh)
			thatValue = CS_KING+1;
		if(thatValue > value)
		{
			value = thatValue;
			card = [cardList objectAt:i];
		}
	}
	
	return card;
}

- sortPile
/*
	
*/
{
	int i;
	int numItems = [cardList count];
	id sortedList = [[List allocFromZone:[self zone]] initCount:numItems];
	
	for(i=0; i<numItems; i++)
	{
		[sortedList addObject:[self lowestCard]];
		[self removeCard:[self lowestCard]];
	}
	
	[[cardList empty] appendList:sortedList];
	
	return self;
}

- sortList:listObject
/*
	Returns a sorted copy of listObject. 
	
	This is method should not be called by clients.  It is meant to be used internally.
*/
{
	int i, lowcard, thisValue, thatValue;
	id sortedList;
	
	// return if there's only 1 card in aPile
	if([listObject count] == 1) return listObject;
	
	// find the lowest card, sort the rest
	lowcard = 0;
	for(i=1; i<[listObject count]; i++)
	{
		thisValue = [[listObject objectAt:i] value];
		if(thisValue == CS_ACE && aceHigh)
			thisValue = CS_KING+1;
		thatValue = [[listObject objectAt:lowcard] value];
		if(thatValue == CS_ACE && aceHigh)
			thatValue = CS_KING+1;
		if(thisValue < thatValue)
			lowcard = i;
	}
	
	sortedList = [[List alloc] init];
	
	[sortedList addObject:[listObject objectAt:lowcard]];
	[listObject removeObject:[listObject objectAt:lowcard]];
	[sortedList appendList:[self sortList:listObject]];
	
	return sortedList;
}

- cardList
{
	return cardList;
}

- (PokerHand)pokerHand
/*
	Returns the highest poker hand found in the Pile, or CS_NOPOKERHAND if none found.
*/
{
	if([self isRoyalFlush])		return CS_ROYALFLUSH;
	if([self isStraightFlush])	return CS_STRAIGHTFLUSH;
	if([self isFourOfKind])		return CS_FOUROFKIND;
	if([self isFullHouse])		return CS_FULLHOUSE;
	if([self isFlush])			return CS_FLUSH;
	if([self isStraight])		return CS_STRAIGHT;
	if([self isThreeOfKind])	return CS_THREEOFKIND;
	if([self isTwoPair])		return CS_TWOPAIR;
	if([self isJacksOrBetter])	return CS_JACKSORBETTER;
	
	return CS_NOPOKERHAND;
}

@end
